/* Copyright (c) 2013, Mohamed Moubarak
 * Copyright (c) 2012, Marios Hadjieleftheriou
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.spatialindex;

public class Point implements IShape, Cloneable {
	public double[] m_pCoords = null;

	public Point(double[] pCoords) {
		m_pCoords = new double[pCoords.length];
		System.arraycopy(pCoords, 0, m_pCoords, 0, pCoords.length);
	}

	public Point(final Point pt) {
		m_pCoords = new double[pt.m_pCoords.length];
		System.arraycopy(pt.m_pCoords, 0, m_pCoords, 0, pt.m_pCoords.length);
	}

	public Point(double lat, double lon) {
		this(new double[] { lat, lon });
	}

	public boolean equals(Object o) {
		if (o instanceof Point) {
			Point pt = (Point) o;

			if (pt.m_pCoords.length != m_pCoords.length)
				return false;

			for (int cIndex = 0; cIndex < m_pCoords.length; cIndex++) {
				if (m_pCoords[cIndex] < pt.m_pCoords[cIndex] - SpatialIndex.EPSILON
				                || m_pCoords[cIndex] > pt.m_pCoords[cIndex] + SpatialIndex.EPSILON)
					return false;
			}

			return true;
		}

		return false;
	}

	//
	// Cloneable interface
	//

	public Object clone() {
		return new Point(m_pCoords);
	}

	//
	// IShape interface
	//

	public boolean intersects(final IShape s) {
		if (s instanceof Region)
			return ((Region) s).contains(this);

		return false;
	}

	public boolean contains(final IShape s) {
		return false;
	}

	public boolean touches(final IShape s) {
		if (s instanceof Point && this.equals(s))
			return true;

		if (s instanceof Region)
			return ((Region) s).touches(this);

		return false;
	}

	public double[] getCenter() {
		double[] pCoords = new double[m_pCoords.length];
		System.arraycopy(m_pCoords, 0, pCoords, 0, m_pCoords.length);
		return pCoords;
	}

	public long getDimension() {
		return m_pCoords.length;
	}

	public Region getMBR() {
		return new Region(m_pCoords, m_pCoords);
	}

	public double getArea() {
		return 0.0f;
	}

	public double getMinimumDistance(final IShape s) {
		if (s instanceof Region)
			return ((Region) s).getMinimumDistance(this);

		if (s instanceof Point)
			return getMinimumDistance((Point) s);

		throw new IllegalStateException("getMinimumDistance: Not implemented yet!");
	}

	double getMinimumDistance(final Point p) {
		if (m_pCoords.length != p.m_pCoords.length)
			throw new IllegalArgumentException("getMinimumDistance: Shape has the wrong number of dimensions.");

		double ret = 0.0;

		for (int cIndex = 0; cIndex < m_pCoords.length; cIndex++) {
			ret += Math.pow(m_pCoords[cIndex] - p.m_pCoords[cIndex], 2.0);
		}

		return Math.sqrt(ret);
	}

	public double getCoord(int index) throws IndexOutOfBoundsException {
		if (index >= m_pCoords.length)
			throw new IndexOutOfBoundsException("" + index);
		return m_pCoords[index];
	}
}
