/* Copyright (c) 2013, Mohamed Moubarak
 * Copyright (c) 2012, Marios Hadjieleftheriou
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.spatialindex.rtree;

import org.spatialindex.*;

import java.util.*;

public class Leaf extends Node {
	public Leaf(RTree pTree, int id) {
		super(pTree, id, 0, pTree.m_leafCapacity);
	}

	protected Node chooseSubtree(Region mbr, int level, Stack<Integer> pathBuffer) {
		return this;
	}

	protected Leaf findLeaf(Region mbr, int id, Stack<Integer> pathBuffer) {
		for (int cChild = 0; cChild < m_children; cChild++) {
			if (m_pIdentifier[cChild] == id && mbr.equals(m_pMBR[cChild]))
				return this;
		}

		return null;
	}

	protected Node[] split(byte[] pData, Region mbr, int id) {
		m_pTree.m_stats.m_splits++;

		ArrayList<Integer> g1 = new ArrayList<Integer>(), g2 = new ArrayList<Integer>();

		switch (m_pTree.m_treeVariant) {
			case SpatialIndex.RtreeVariantLinear:
			case SpatialIndex.RtreeVariantQuadratic:
				rtreeSplit(pData, mbr, id, g1, g2);
				break;
			case SpatialIndex.RtreeVariantRstar:
				rstarSplit(pData, mbr, id, g1, g2);
				break;
			default:
				throw new IllegalStateException("Unknown RTree variant.");
		}

		Node left = new Leaf(m_pTree, -1);
		Node right = new Leaf(m_pTree, -1);

		int cIndex;

		for (cIndex = 0; cIndex < g1.size(); cIndex++) {
			int i = g1.get(cIndex).intValue();
			left.insertEntry(m_pData[i], m_pMBR[i], m_pIdentifier[i]);

			// we don't want to delete the data array from this node's destructor!
			m_pData[i] = null;
		}

		for (cIndex = 0; cIndex < g2.size(); cIndex++) {
			int i = g2.get(cIndex).intValue();
			right.insertEntry(m_pData[i], m_pMBR[i], m_pIdentifier[i]);

			// we don't want to delete the data array from this node's destructor!
			m_pData[i] = null;
		}

		Node[] ret = new Node[2];
		ret[0] = left;
		ret[1] = right;
		return ret;
	}

	protected void deleteData(int id, Stack<Integer> pathBuffer) {
		int child;
		for (child = 0; child < m_children; child++) {
			if (m_pIdentifier[child] == id)
				break;
		}

		deleteEntry(child);
		m_pTree.writeNode(this);

		Stack<Node> toReinsert = new Stack<Node>();
		condenseTree(toReinsert, pathBuffer);

		// re-insert eliminated nodes.
		while (!toReinsert.empty()) {
			Node n = (Node) toReinsert.pop();
			m_pTree.deleteNode(n);

			for (int cChild = 0; cChild < n.m_children; cChild++) {
				// keep this in the for loop. The tree height might change after insertions.
				boolean[] overflowTable = new boolean[m_pTree.m_stats.m_treeHeight];
				for (int cLevel = 0; cLevel < m_pTree.m_stats.m_treeHeight; cLevel++)
					overflowTable[cLevel] = false;

				m_pTree.insertData_impl(n.m_pData[cChild], n.m_pMBR[cChild], n.m_pIdentifier[cChild], n.m_level,
				                overflowTable);
				n.m_pData[cChild] = null;
			}
		}
	}
}
