/* Copyright (c) 2013, Mohamed Moubarak
 * Copyright (c) 2012, Marios Hadjieleftheriou
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.spatialindex.storagemanager;

import java.util.*;
import java.io.*;

public class DiskStorageManager implements IStorageManager {
	private RandomAccessFile m_dataFile = null;
	private RandomAccessFile m_indexFile = null;
	private int m_pageSize = 0;
	private int m_nextPage = -1;
	private TreeSet<Integer> m_emptyPages = new TreeSet<Integer>();
	private HashMap<Integer, Entry> m_pageIndex = new HashMap<Integer, Entry>();
	private byte[] m_buffer = null;

	public DiskStorageManager(String name, int pageSize) throws SecurityException, NullPointerException, IOException,
	                FileNotFoundException, IllegalArgumentException {
		if (name == null)
			throw new IllegalArgumentException("Property FileName was not specified.");
		initFiles(name, false);
		m_pageSize = pageSize;
		m_nextPage = 0;
		m_buffer = new byte[m_pageSize];
	}

	public DiskStorageManager(String name) throws SecurityException, NullPointerException, IOException,
	                FileNotFoundException, IllegalArgumentException {
		if (name == null)
			throw new IllegalArgumentException("Property FileName was not specified.");
		initFiles(name, true);

		try {
			m_pageSize = m_indexFile.readInt();
		}
		catch (EOFException ex) {
			throw new IllegalStateException("Failed reading pageSize.");
		}

		try {
			m_nextPage = m_indexFile.readInt();
		}
		catch (EOFException ex) {
			throw new IllegalStateException("Failed reading nextPage.");
		}

		// create buffer.
		m_buffer = new byte[m_pageSize];

		try {
			int count = m_indexFile.readInt();

			// load empty pages in memory
			for (int cCount = 0; cCount < count; cCount++)
				m_emptyPages.add(m_indexFile.readInt());

			// load index table in memory.
			count = m_indexFile.readInt();
			for (int cCount = 0; cCount < count; cCount++) {
				Entry e = new Entry();
				int id = m_indexFile.readInt();
				e.m_length = m_indexFile.readInt();

				int count2 = m_indexFile.readInt();

				for (int cCount2 = 0; cCount2 < count2; cCount2++) {
					e.m_pages.add(m_indexFile.readInt());
				}
				m_pageIndex.put(id, e);
			}
		}
		catch (EOFException ex) {
			throw new IllegalStateException("Corrupted index file.");
		}

	}

	private void initFiles(String name, boolean reuse) throws IOException {
		File indexFile = new File(name + ".idx");
		File dataFile = new File(name + ".dat");
		if (!reuse) {
			if (indexFile.exists())
				indexFile.delete();
			if (dataFile.exists())
				dataFile.delete();
			if (!indexFile.createNewFile())
				throw new IOException("Index file cannot be opened: " + indexFile.getAbsolutePath());
			if (!dataFile.createNewFile())
				throw new IOException("Data file cannot be opened: " + dataFile.getAbsolutePath());
		}
		m_indexFile = new RandomAccessFile(indexFile, "rw");
		m_dataFile = new RandomAccessFile(dataFile, "rw");
	}

	public int entries() {
		return m_pageIndex.size();
	}

	public void flush() {
		try {
			m_indexFile.seek(0l);
			m_indexFile.writeInt(m_pageSize);
			m_indexFile.writeInt(m_nextPage);
			m_indexFile.writeInt(m_emptyPages.size());

			for (int p : m_emptyPages)
				m_indexFile.writeInt(p);

			m_indexFile.writeInt(m_pageIndex.size());

			for (Map.Entry<Integer, Entry> me : m_pageIndex.entrySet()) {
				m_indexFile.writeInt(me.getKey());
				Entry e = me.getValue();
				m_indexFile.writeInt(e.m_length);
				m_indexFile.writeInt(e.m_pages.size());
				for (int page : e.m_pages)
					m_indexFile.writeInt(page);
			}
		}
		catch (IOException ex) {
			throw new IllegalStateException("Corrupted index file.");
		}
		finally {
			close();
		}
	}

	public byte[] loadByteArray(final int id) {
		if (!m_pageIndex.containsKey(id))
			throw new InvalidPageException(id);
		return loadByteArray(m_pageIndex.get(id));
	}

	private byte[] loadByteArray(Entry e) {
		byte[] data = new byte[e.m_length];
		int cIndex = 0;
		int cRem = e.m_length;

		for (int page : e.m_pages) {
			try {
				m_dataFile.seek(page * m_pageSize);
				if (m_dataFile.read(m_buffer) != m_pageSize)
					throw new IllegalStateException("Corrupted data file.");
			}
			catch (IOException ex) {
				throw new IllegalStateException("Corrupted data file.");
			}
			int cLen = (cRem > m_pageSize) ? m_pageSize : cRem;
			System.arraycopy(m_buffer, 0, data, cIndex, cLen);
			cIndex += cLen;
			cRem -= cLen;
		}
		return data;
	}

	public int storeByteArray(final int id, final byte[] data) {
		if (id == NewPage) {
			Entry e = new Entry();
			e.m_length = data.length;

			int cIndex = 0;
			int cPage;
			int cRem = data.length;
			int cLen;

			while (cRem > 0) {
				if (!m_emptyPages.isEmpty()) {
					Integer i = (Integer) m_emptyPages.first();
					m_emptyPages.remove(i);
					cPage = i.intValue();
				}
				else {
					cPage = m_nextPage;
					m_nextPage++;
				}

				cLen = (cRem > m_pageSize) ? m_pageSize : cRem;
				System.arraycopy(data, cIndex, m_buffer, 0, cLen);

				try {
					m_dataFile.seek(cPage * m_pageSize);
					m_dataFile.write(m_buffer);
				}
				catch (IOException ex) {
					throw new IllegalStateException("Corrupted data file.");
				}

				cIndex += cLen;
				cRem -= cLen;
				e.m_pages.add(new Integer(cPage));
			}

			int i = e.m_pages.get(0);
			m_pageIndex.put(e.m_pages.get(0), e);
			return i;
		}
		else {
			// find the entry.
			Entry oldEntry = m_pageIndex.get(id);
			if (oldEntry == null)
				throw new InvalidPageException(id);

			m_pageIndex.remove(new Integer(id));

			Entry e = new Entry();
			e.m_length = data.length;

			int cIndex = 0;
			int cPage;
			int cRem = data.length;
			int cLen, cNext = 0;

			while (cRem > 0) {
				if (cNext < oldEntry.m_pages.size()) {
					cPage = oldEntry.m_pages.get(cNext);
					cNext++;
				}
				else if (!m_emptyPages.isEmpty()) {
					int i = m_emptyPages.first();
					m_emptyPages.remove(i);
					cPage = i;
				}
				else {
					cPage = m_nextPage;
					m_nextPage++;
				}

				cLen = (cRem > m_pageSize) ? m_pageSize : cRem;
				System.arraycopy(data, cIndex, m_buffer, 0, cLen);

				try {
					m_dataFile.seek(cPage * m_pageSize);
					m_dataFile.write(m_buffer);
				}
				catch (IOException ex) {
					throw new IllegalStateException("Corrupted data file.");
				}

				cIndex += cLen;
				cRem -= cLen;
				e.m_pages.add(new Integer(cPage));
			}

			while (cNext < oldEntry.m_pages.size()) {
				m_emptyPages.add(oldEntry.m_pages.get(cNext));
				cNext++;
			}

			int i = e.m_pages.get(0);
			m_pageIndex.put(i, e);
			return i;
		}
	}

	public void deleteByteArray(final int id) {
		// find the entry.
		Entry e = m_pageIndex.get(id);
		if (e == null)
			throw new InvalidPageException(id);
		m_pageIndex.remove(id);
		for (int cIndex = 0; cIndex < e.m_pages.size(); cIndex++)
			m_emptyPages.add(e.m_pages.get(cIndex));
	}

	public void close() {
		try {
			m_dataFile.close();
		}
		catch (IOException e) {
		}
		try {
			m_indexFile.close();
		}
		catch (IOException e) {
		}
	}

	class Entry {
		int m_length = 0;
		ArrayList<Integer> m_pages = new ArrayList<Integer>();

		int page(int i) {
			return m_pages.get(i);
		}
	}
}
